unit Character;

interface

uses
  World, Graphics;

type
  TCharacterStates = set of (csGoLeft, csGoRight, csLeft,
                             csRight, csJump, csFall);

  TCharacter = class
  private
    FWorld: TWorld;
    FPicture: TPicture;
    FXPos, FYPos: Extended;
    FState: TCharacterStates;
    FWidth, FHeight: Integer;
    FSpeed, FMaxSpeed, FSpeedInc, FFrictionModifier: Extended;
    FJump, FJumpImpulse, FMaxFallSpeed, FGravityModifier: Extended;
    function GetX: Integer;
    function GetY: Integer;
    procedure SetState(const Value: TCharacterStates);
    procedure CheckDown;
    procedure CheckUp;
    procedure CheckLeft;
    procedure CheckRight;
  public
    constructor Create(aWorld: TWorld; aPicture: TPicture;
                       aX, aY: Integer;
                       aMaxSpeed: Extended = 10;
                       aSpeedInc: Extended = 2;
                       aJumpImpulse: Extended = 12;
                       aMaxFallSpeed: Extended = 5;
                       aFrictionModifier: Extended = 1;
                       aGravityModifier: Extended = 1);
    procedure GoLeft;
    procedure GoRight;
    procedure Stop;
    procedure Jump;
    procedure Fall;
    procedure Tick;
    procedure Draw(Canvas: TCanvas; aX, aY: Integer);
    property X: Integer read GetX;
    property Y: Integer read GetY;
    property State: TCharacterStates read FState write SetState;
  end;

implementation

{ TCharacter }

procedure TCharacter.CheckDown;
var
  sX, sY, eX: Integer;
begin
  sX := Trunc(FXPos);
  sY := Trunc(FYPos + FHeight);
  eX := Trunc(FXPos + FWidth - 1);
  if FWorld.PixelLevel[sX,sY].Solid or
     FWorld.PixelLevel[eX,sY].Solid then
  begin
    FYPos := (sY div FWorld.TileHeight) * FWorld.TileHeight
      - FHeight;
    FState := FState - [csFall];
    FJump := 0;
  end;
end;

procedure TCharacter.CheckLeft;
var sX, sY, eY: Integer;
begin
  sX := Trunc(FXPos);
  sY := Trunc(FYPos);
  eY := Trunc(FYPos + FHeight - 1);
  if FWorld.PixelLevel[sX,sY].Solid or
     FWorld.PixelLevel[sX,eY].Solid then
  begin
    FState := FState - [csGoLeft, csGoRight] + [csLeft];
    FXPos := (sX div FWorld.TileWidth) * FWorld.TileWidth + FWorld.TileWidth;
    FSpeed := 0;
  end;
  if not (csJump in FState) then Fall;
end;

procedure TCharacter.CheckRight;
var sX, sY, eY: Integer;
begin
  sX := Trunc(FXPos + FWidth);
  sY := Trunc(FYPos);
  eY := Trunc(FYPos + FHeight - 1);
  if FWorld.PixelLevel[sX,sY].Solid or
     FWorld.PixelLevel[sX,eY].Solid then
  begin
    FState := FState - [csGoLeft, csGoRight] + [csRight];
    FXPos := (sX div FWorld.TileWidth) * FWorld.TileWidth - FWorld.TileWidth;
    FSpeed := 0;
  end;
  if not (csJump in FState) then Fall;
end;

procedure TCharacter.CheckUp;
var
  sX, sY, eX: Integer;
begin
  sX := Trunc(FXPos);
  sY := Trunc(FYPos);
  eX := Trunc(FXPos + FWidth - 1);
  if FWorld.PixelLevel[sX,sY].Solid or
     FWorld.PixelLevel[eX,sY].Solid then
  begin
    FYPos := (sY div FWorld.TileHeight) * FWorld.TileHeight + FWorld.TileHeight;
    FJump := 0;
    Fall;
  end;
end;

constructor TCharacter.Create(aWorld: TWorld; aPicture: TPicture; aX,
  aY: Integer; aMaxSpeed, aSpeedInc, aJumpImpulse, aMaxFallSpeed,
  aFrictionModifier, aGravityModifier: Extended);
begin
  FWorld := aWorld;
  FPicture := aPicture;
  FWidth := aPicture.Width;
  FHeight := aPicture.Height;
  FXPos := aX;
  FYPos := aY;
  FMaxSpeed := aMaxSpeed;
  FSpeedInc := aSpeedInc;
  FJumpImpulse := aJumpImpulse;
  FMaxFallSpeed := aMaxFallSpeed;
  FFrictionModifier := aFrictionModifier;
  FGravityModifier := aGravityModifier;
  Fall;
end;

procedure TCharacter.Draw(Canvas: TCanvas; aX, aY: Integer);
begin
  Canvas.Draw(aX + X, aY + Y, FPicture.Graphic);
end;

procedure TCharacter.Fall;
var
  sX, sY, eX: Integer;
begin
  sX := Trunc(FXPos);
  sY := Trunc(FYPos + FHeight);
  eX := Trunc(FXPos + FWidth - 1);
  if ((not FWorld.PixelLevel[sX,sY].Solid) and
     (not FWorld.PixelLevel[eX,sY].Solid)) or (csJump in FState) then
  begin
    FState := FState - [csJump] + [csFall];
  end;
end;

function TCharacter.GetX: Integer;
begin
  Result := Trunc(FXPos);
end;

function TCharacter.GetY: Integer;
begin
  Result := Trunc(FYPos);
end;

procedure TCharacter.GoLeft;
begin
  FState := FState - [csLeft, csRight, csGoRight] + [csGoLeft];
end;

procedure TCharacter.GoRight;
begin
  FState := FState - [csLeft, csRight, csGoLeft] + [csGoRight];
end;

procedure TCharacter.Jump;
begin
  if (not (csFall in FState)) and (not (csJump in FState)) then
  begin
    FState := FState + [csJump];
    FJump := FJumpImpulse;
    FJump := FJump;
  end;
end;

procedure TCharacter.SetState(const Value: TCharacterStates);
begin
  FState := Value;
end;

procedure TCharacter.Stop;
begin
  if csGoLeft in FState then
    FState := FState - [csGoLeft] + [csLeft]
  else if csGoRight in FState then
    FState := FState - [csGoRight] + [csRight];
end;

procedure TCharacter.Tick;
var
  Friction: Extended;
begin
  Friction := FWorld.PixelLevel[(2 * X + FWidth) div 2, Y + FHeight + 1].Friction;

  if csGoLeft in FState then
  begin
    FSpeed := FSpeed - FSpeedInc * Friction * FFrictionModifier;
    if FSpeed < -FMaxSpeed then FSpeed := -FMaxSpeed;
  end;

  if csGoRight in FState then
  begin
    FSpeed := FSpeed + FSpeedInc * Friction * FFrictionModifier;
    if FSpeed > FMaxSpeed then FSpeed := FMaxSpeed;
  end;

  if (csLeft in FState) or (csRight in FState) then
  begin
    if FSpeed > 0 then
    begin
      FSpeed := FSpeed - Friction * FFrictionModifier;
      if FSpeed < 0 then FSpeed := 0;
    end;
    if FSpeed < 0 then
    begin
      FSpeed := FSpeed + Friction * FFrictionModifier;
      if FSpeed > 0 then FSpeed := 0;
    end;
  end;

  if csJump in FState then
  begin
    FJump := FJump - FWorld.Gravity * FGravityModifier;
    if FJump < 0 then Fall;
  end;

  if csFall in FState then
  begin
    FJump := FJump - FWorld.Gravity * FGravityModifier;
    if FJump > FMaxFallSpeed then FJump := FMaxFallSpeed;
  end;

  FXPos := FXPos + FSpeed;
  FYPos := FYPos - FJump;

  if Abs(FJump) > Abs(FSpeed) then
  begin
    if FJump > 0 then CheckUp;
    if FJump < 0 then CheckDown;
    if FSpeed > 0 then CheckRight;
    if FSpeed < 0 then CheckLeft;
  end
  else
  begin
    if FSpeed > 0 then CheckRight;
    if FSpeed < 0 then CheckLeft;
    if FJump > 0 then CheckUp;
    if FJump < 0 then CheckDown;
  end;
end;

end.
